import os
import sys
import argparse
import inspect
from ..exporters import exporters


def _get_caller():
    return inspect.getmodule(inspect.currentframe().f_back.f_back)


def _get_parser(caller, basename, path, formats, verbosity):
    if basename is None:
        basename = os.path.splitext(os.path.basename(caller.__file__))[0]

    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--name",
        "-n",
        default=basename,
        help="Output basename",
    )
    parser.add_argument(
        "--path",
        default=path,
        help="Output path",
    )
    parser.add_argument(
        "--formats", "-f",
        nargs="+",
        choices=list(sum((e.extensions for e in exporters), [])),
        default=formats,
        help="Formates to render",
        metavar="format"
    )
    parser.add_argument(
        "--verbosity", "-v",
        type=int,
        default=int(verbosity)
    )

    exporters.set_options(parser)

    return parser


def get_parser(basename=None, path="/tmp", formats=["html"], verbosity=1):
    caller = _get_caller()
    return _get_parser(caller, basename, path, formats, verbosity)


def run(animation, ns):
    if ns.name == "-" and len(ns.formats) == 1:
        file = sys.stdout.buffer if ns.formats[0] == "tgs" else sys.stdout
        exporter = exporters.get_from_extension(ns.formats[0])
        exporter.export(animation, absname, exporter.argparse_options(ns))
    else:
        for fmt in ns.formats:
            absname = os.path.abspath(os.path.join(ns.path, ns.name + "." + fmt))
            if ns.verbosity:
                print("file://" + absname)
            exporter = exporters.get_from_extension(fmt)
            exporter.export(animation, absname, exporter.argparse_options(ns))


def script_main(animation, basename=None, path="/tmp", formats=["html"], verbosity=1):
    """
    Sets up a script to output an animation into various formats
    """
    caller = _get_caller()
    if caller.__name__ == "__main__":
        parser = _get_parser(caller, basename, path, formats, verbosity)
        run(animation, parser.parse_args())



# NOTE: requires pillow, pypotrace>=0.2, numpy, scipy to be installed
from PIL import Image
import potrace
import numpy
import enum
from scipy.cluster.vq import kmeans
from .. import objects
from ..nvector import NVector


class QuanzationMode(enum.Enum):
    Nearest = 1
    Exact = 2


class RasterImage:
    def __init__(self, data):
        self.data = data

    @classmethod
    def from_pil(cls, image):
        return cls(numpy.array(image))

    @classmethod
    def open(cls, filename):
        return cls.from_pil(Image.open(filename))

    def k_means(self, n_colors):
        """!
        Returns a list of centroids
        """
        colors = []
        for row in range(self.data.shape[0]):
            for column in range(self.data.shape[1]):
                if self.get_alpha(row, column) == 255:
                    colors.append(self.data[row][column])

        colors = numpy.array(colors, numpy.float)
        return kmeans(colors, n_colors+1)[0]

    def get_alpha(self, row, column):
        if self.data.shape[2] >= 4:
            return self.data[row][column][3]
        return 255

    def quantize(self, codebook, quantization_mode=QuanzationMode.Nearest):
        """!
        Returns a list of tuple [color, data] where for each color in codebook
        data is a bit mask for the image

        You can get codebook from k_means
        """
        if codebook is None or len(codebook) == 0:
            return [(numpy.array([0., 0., 0., 255.]), self.mono())]

        mono_data = []
        for c in codebook:
            mono_data.append((c, numpy.zeros(self.data.shape[:2])))

        for row in range(self.data.shape[0]):
            for column in range(self.data.shape[1]):
                if self.get_alpha(row, column) == 255:
                    if quantization_mode == QuanzationMode.Nearest:
                        min_norm = 511 # (norm of [255, 255, 255, 255]) + 1
                        best = None
                        for color, bitmap in mono_data:
                            norm = numpy.linalg.norm(self.data[row][column] - color)
                            if norm < min_norm:
                                min_norm = norm
                                best = bitmap
                                if norm == 0:
                                    break
                        best[row][column] = 1
                    else:
                        for color, bitmap in mono_data:
                            if numpy.array_equal(color, self.data[row][column]):
                                bitmap[row][column] = 1
                                break

        return mono_data

    def mono(self):
        """!
        Returns a bit mask of opaque pixels
        """
        mono_data = numpy.zeros(self.data.shape[:2])
        for row in range(self.data.shape[0]):
            for column in range(self.data.shape[1]):
                mono_data[row][column] = int(self.data[row][column][3] == 255)
        return mono_data


class Vectorizer:
    def __init__(self, *a, **kw):
        self.animation = objects.Animation(*a, **kw)
        self.palette = None
        self.layers = {}

    def _create_layer(self, layer_name):
        layer = self.animation.add_layer(objects.ShapeLayer())
        if layer_name:
            self.layers[layer_name] = layer
            layer.name = layer_name
        return layer

    def prepare_layer(self, layer_name=None):
        layer = self._create_layer(layer_name)
        layer._max_verts = {}
        if self.palette is None:
            group = layer.add_shape(objects.Group())
            group.name = "bitmap"
            layer._max_verts[group.name] = 0
            group.add_shape(objects.Path())
            group.add_shape(objects.Fill(NVector(0, 0, 0)))
        else:
            for color in self.palette:
                group = layer.add_shape(objects.Group())
                group.name = "color_%s" % "".join("%02x" % int(c) for c in color)
                layer._max_verts[group.name] = 0
                fcol = color/255
                fill = group.add_shape(objects.Fill(NVector(*fcol)))
                if len(fcol) > 3 and fcol[3] < 1:
                    fill.opacity.value = fcol[3] * 100
        return layer

    def raster_to_layer(self, raster, layer_name=None, mode=QuanzationMode.Nearest):
        layer = self.prepare_layer()
        mono_data = raster.quantize(self.palette, mode)
        for (color, bitmap), group in zip(mono_data, layer.shapes):
            self.raster_to_bezier(group, bitmap)

    def raster_to_bezier(self, group, mono_data):
        bmp = potrace.Bitmap(mono_data)
        path = bmp.trace()
        shapes = []
        for curve in path:
            shape = group.insert_shape(0, objects.Path())
            shapes.append(shape)
            bezier = shape.shape.value
            bezier.add_point(NVector(*curve.start_point))
            for segment in curve:
                if segment.is_corner:
                    bezier.add_point(NVector(*segment.c))
                    bezier.add_point(NVector(*segment.end_point))
                else:
                    sp = NVector(*bezier.vertices[-1])
                    ep = NVector(*segment.end_point)
                    c1 = NVector(*segment.c1) - sp
                    c2 = NVector(*segment.c2) - ep
                    bezier.out_tangents[-1] = c1
                    bezier.add_point(ep, c2)
        return shapes

    def raster_to_frame(self, raster, layer_name, time, mode=QuanzationMode.Nearest):
        if layer_name not in self.layers:
            layer = self.prepare_layer(layer_name)
        else:
            layer = self.layers[layer_name]

        mono_data = raster.quantize(self.palette, mode)
        for (color, bitmap), group in zip(mono_data, layer.shapes):
            shapes = self.raster_to_bezier(group, bitmap)
            if shapes:
                # TODO handle multiple shapes
                nverts = len(shapes[0].shape.value.vertices)
                if nverts > layer._max_verts[group.name]:
                    layer._max_verts[group.name] = nverts
                for shape in shapes:
                    bezier = shape.shape.value
                    shape.shape.add_keyframe(time, bezier)

    def adjust_missing_vertices(self, layer_name):
        layer = self.layers[layer_name]
        for group in layer.shapes:
            # TODO handle multiple shapes
            shape = group.shapes[0]
            nverts = layer._max_verts[group.name]
            for kf in shape.shape.keyframes:
                bezier = kf.start
                count = nverts - len(bezier.vertices)
                bezier.vertices += [bezier.vertices[-1]] * count
                bezier.in_tangents += [NVector(0, 0)] * count
                bezier.out_tangents += [NVector(0, 0)] * count

    def duplicate_start_frame(self, layer_name, time):
        layer = self.layers[layer_name]
        for group in layer.shapes:
            shape = group.shapes[0]
            bezier = shape.shape.keyframes[0].start
            group.shapes[0].shape.add_keyframe(time, bezier)


def raster_to_animation(filename, n_colors=1, palette=[], mode=QuanzationMode.Nearest):
    vc = Vectorizer(60)
    raster = RasterImage.open(filename)
    if palette:
        vc.palette = palette
    elif n_colors > 1:
        vc.palette = raster.k_means(n_colors)

    vc.raster_to_layer(raster, None, mode)
    return vc.animation


def color2numpy(vcolor):
    l = (vcolor * 255).components
    if len(l) == 3:
        l.append(255)
    return numpy.array(l, numpy.uint8)


def raster_frames_to_animation(files, n_colors=1, frame_delay=1, start_time=0,
                               looping=True, framerate=60, palette=[], mode=QuanzationMode.Nearest):
    nframes = len(files)
    if not looping:
        nframes -= 1
    vc = Vectorizer(start_time + frame_delay * nframes, framerate=framerate)

    raster = RasterImage.open(files.pop(0))
    if palette:
        vc.palette = [color2numpy(c) for c in palette]
    elif n_colors > 1:
        vc.palette = raster.k_means(n_colors)

    time = start_time
    vc.raster_to_frame(raster, "anim", time, mode)

    for filename in files:
        raster = RasterImage.open(filename)
        time += frame_delay
        vc.raster_to_frame(raster, "anim", time, mode)

    vc.adjust_missing_vertices("anim")
    if looping:
        vc.duplicate_start_frame("anim", time + frame_delay)

    return vc.animation



from .base import TgsObject, TgsProp, PseudoBool, TgsEnum
from .effects import Effect
from .helpers import Transform, Mask
from .shapes import ShapeElement
from .text import TextAnimatorData
from .properties import Value


## \ingroup Lottie
class BlendMode(TgsEnum):
    Normal = 0
    Multiply = 1
    Screen = 2
    Overlay = 3
    Darken = 4
    Lighten = 5
    ColorDodge = 6
    ColorBurn = 7
    HardLight = 8
    SoftLight = 9
    Difference = 10
    Exclusion = 11
    Hue = 12
    Saturation = 13
    Color = 14
    Luminosity = 15


## \ingroup Lottie
## \todo SVG masks, transparent gradients
class MatteMode(TgsEnum):
    Normal = 0
    Alpha = 1
    InvertedAlpha = 2
    Luma = 3
    InvertedLuma = 4


## \ingroup Lottie
class Layer(TgsObject):
    _props = [
        TgsProp("type", "ty", int, False),
        TgsProp("transform", "ks", Transform, False),
        TgsProp("auto_orient", "ao", PseudoBool, False),
        TgsProp("blend_mode", "bm", BlendMode, False),
        TgsProp("matte_mode", "tt", MatteMode, False),
        TgsProp("threedimensional", "ddd", PseudoBool, False),
        TgsProp("index", "ind", int, False),
        #TgsProp("css_class", "cl", str, False),
        #TgsProp("layer_html_id", "ln", str, False),
        TgsProp("in_point", "ip", float, False),
        TgsProp("out_point", "op", float, False),
        TgsProp("start_time", "st", float, False),
        TgsProp("name", "nm", str, False),
        TgsProp("has_masks", "hasMask", bool, False),
        TgsProp("masks", "masksProperties", Mask, True),
        TgsProp("effects", "ef", Effect, True),
        TgsProp("stretch", "sr", float, False),
        TgsProp("parent", "parent", int, False),
    ]
    ## %Layer type.
    ## @see https://github.com/bodymovin/bodymovin-extension/blob/master/bundle/jsx/enums/layerTypes.jsx
    type = None
    _classses = {}

    @property
    def has_masks(self):
        """!
        Whether the layer has some masks applied
        """
        return bool(self.masks) if getattr(self, "masks") is not None else None

    def __init__(self):
        ## Transform properties
        self.transform = Transform()
        ## Auto-Orient along path AE property.
        self.auto_orient = False
        ## 3d layer flag
        self.threedimensional = False
        ## Layer index in AE. Used for parenting and expressions.
        self.index = None

        """
        # Parsed layer name used as html class on SVG/HTML renderer
        #self.css_class = ""
        # Parsed layer name used as html id on SVG/HTML renderer
        #self.layer_html_id = ""
        """
        ## In Point of layer. Sets the initial frame of the layer.
        self.in_point = None
        ## Out Point of layer. Sets the final frame of the layer.
        self.out_point = None
        ## Start Time of layer. Sets the start time of the layer.
        self.start_time = 0
        ## After Effects Layer Name. Used for expressions.
        self.name = None
        ## List of Effects
        self.effects = None
        ## Layer Time Stretching
        self.stretch = 1
        ## Layer Parent. Uses ind of parent.
        self.parent = None
        ## List of Masks
        self.masks = None
        ## Blend Mode
        self.blend_mode = BlendMode.Normal
        ## Matte mode, the layer will inherit the transparency from the layer above
        self.matte_mode = None

    @classmethod
    def _load_get_class(cls, lottiedict):
        if not Layer._classses:
            Layer._classses = {
                sc.type: sc
                for sc in Layer.__subclasses__()
            }
        return Layer._classses[lottiedict["ty"]]


## \ingroup Lottie
class NullLayer(Layer):
    """!
    Layer with no data, useful to group layers together
    """
    ## %Layer type.
    type = 3

    def __init__(self):
        Layer.__init__(self)


## \ingroup Lottie
class TextLayer(Layer):
    _props = [
        TgsProp("data", "t", TextAnimatorData, False),
    ]
    ## %Layer type.
    type = 5

    def __init__(self):
        Layer.__init__(self)
        ## Text Data
        self.data = TextAnimatorData()


## \ingroup Lottie
class ShapeLayer(Layer):
    """!
    Layer containing ShapeElement objects
    """
    _props = [
        TgsProp("shapes", "shapes", ShapeElement, True),
    ]
    ## %Layer type.
    type = 4

    def __init__(self):
        Layer.__init__(self)
        ## Shape list of items
        self.shapes = [] # ShapeElement

    def add_shape(self, shape):
        self.shapes.append(shape)
        return shape

    def insert_shape(self, index, shape):
        self.shapes.insert(index, shape)
        return shape


## \ingroup Lottie
## \todo SVG/SIF I/O
## \todo importers for raster images without vectorization
class ImageLayer(Layer):
    _props = [
        TgsProp("image_id", "refId", str, False),
    ]
    ## %Layer type.
    type = 2

    def __init__(self, image_id=""):
        Layer.__init__(self)
        ## id pointing to the source image defined on 'assets' object
        self.image_id = image_id


## \ingroup Lottie
## \ingroup LottieCheck
class PreCompLayer(Layer):
    _props = [
        TgsProp("reference_id", "refId", str, False),
        TgsProp("time_remapping", "tm", float, False),
    ]
    ## %Layer type.
    type = 0

    def __init__(self):
        Layer.__init__(self)
        ## id pointing to the source composition defined on 'assets' object
        self.reference_id = ""
        ## Comp's Time remapping
        self.time_remapping = Value()


## \ingroup Lottie
class SolidColorLayer(Layer):
    """!
    Layer with a solid color rectangle
    """
    _props = [
        TgsProp("color", "sc", str, False),
        TgsProp("height", "sh", float, False),
        TgsProp("width", "sw", float, False),
    ]
    ## %Layer type.
    type = 1

    def __init__(self, color="", width=0, height=0):
        Layer.__init__(self)
        ## Color of the layer as a \c \#rrggbb hex
        # \todo Convert NVector to string
        self.color = color
        ## Height of the layer.
        self.height = height
        ## Width of the layer.
        self.width = width



#!/usr/bin/env python3

import sys
import os
import argparse
sys.path.append(os.path.join(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__))),
    "lib"
))
from tgs.parsers.svg.importer import parse_color


parser = argparse.ArgumentParser(
    description="Converts a CSS color into a normalized array, as used in lottie"
)
parser.add_argument(
    "color",
    help="Color to inspect (in one of the CSS color formats)"
)

if __name__ == "__main__":
    ns = parser.parse_args()
    print(parse_color(ns.color).components)




from ..objects.base import TgsObject
from ..objects.bezier import Bezier
from..nvector import NVector


def strip(tgs_object):
    if isinstance(tgs_object, Bezier):
        tgs_object.shape = [NVector(x.x, x.y) for x in tgs_object.shape]
        tgs_object.in_tangents = [NVector(x.x, x.y) for x in tgs_object.in_tangents]
        tgs_object.out_tangents = [NVector(x.x, x.y) for x in tgs_object.out_tangents]

    for p in tgs_object._props:
        pval = p.get(tgs_object)
        if isinstance(pval, TgsObject):
            strip(pval)
        elif isinstance(pval, list) and pval and isinstance(pval[0], TgsObject):
            for c in pval:
                strip(c)
        elif p.lottie in {"ind", "ix", "nm", "mn"}:
            p.set(tgs_object, None)



import json
import gzip
from ..objects import Animation


def parse_tgs_json(filename):
    """!
    Reads both tgs and lottie files, returns the json structure
    """
    with open(filename, "r") as file:
        mn = file.buffer.read(2)
        file.buffer.seek(0)
        if mn == b'\x1f\x8b': # gzip magic number
            json_file = gzip.open(file.buffer, "rb")
        else:
            json_file = file
        return json.load(json_file)


def parse_tgs(filename):
    """!
    Reads both tgs and lottie files
    """
    lottie = parse_tgs_json(filename)
    return Animation.load(lottie)
